const Cc = Components.classes;
const Ci = Components.interfaces;

Components.utils.import("resource://gre/modules/XPCOMUtils.jsm");

// Other constants for us
const OBSERVING = ["xul-window-visible", "domwindowopened", "domwindowclosed",
                   "quit-application-granted"];

// ********************** DEBUGGING *******************************************
function log(aMsg) {
  aMsg = ("\tNetworkPrioritizer: " + aMsg + "\n")
  dump(aMsg);
  Cc["@mozilla.org/consoleservice;1"].getService(Ci.nsIConsoleService)
                                     .logStringMessage(aMsg);
}
// ********************** DEBUGGING *******************************************



const PRIORITIES = {
  focused: {
    active:   Ci.nsISupportsPriority.PRIORITY_HIGHEST,
    inactive: Ci.nsISupportsPriority.PRIORITY_NORMAL },
  unfocused: {
    active:   Ci.nsISupportsPriority.PRIORITY_NORMAL,
    inactive: Ci.nsISupportsPriority.PRIORITY_LOW },
  minimized: {
    active:   Ci.nsISupportsPriority.PRIORITY_LOW,
    inactive: Ci.nsISupportsPriority.PRIORITY_LOWEST }
};


function NetworkPrioritizer() {}

NetworkPrioritizer.prototype = {
  classDescription: "Network Prioritizer",
  contractID: "@mozilla.org/browser/networkprioritizer;1",
  classID: Components.ID("{b8d59546-8dd2-11de-9a6f-1bf8a735809e}"),
  QueryInterface: XPCOMUtils.generateQI([
                                         Ci.nsIDOMEventListener,
                                         Ci.nsIObserver,
                                         Ci.nsISupportsWeakReference]),

  _xpcom_categories: [{category: "app-startup", service: true}],

  _windowWatcher: Cc["@mozilla.org/embedcomp/window-watcher;1"].
                  getService(Ci.nsIWindowWatcher),

  __observerService: null,
  get _observerService() {
    if (!this.__observerService)
      this.__observerService = Cc["@mozilla.org/observer-service;1"].
                               getService(Ci.nsIObserverService);
    return this.__observerService;
  },

  // ["winid": { priority: 1, 
  _windows: [],


  _lastFocusedWindowID: null,

  init: function() {
    log("OMG CHANGE\n");
    // do stuff
    // attach our observers
    let _this = this;
    Array.forEach(OBSERVING, function(topic) {
      _this._observerService.addObserver(_this, topic, true);
    });
  },

  observe: function(aSubject, aTopic, aData) {
    log(aTopic);
    let _this = this;
    switch (aTopic) {
      case "app-startup":
        // this._observerService.removeObserver(this, aTopic, true);
        this.init();
        break;
      case "quit-application-granted":
        this._observerService.removeObserver(this, aTopic, true);
        this.uninit();
        break;
      case "domwindowopened":
        aSubject.addEventListener("load", function(aEvent) {
          aEvent.currentTarget.removeEventListener("load", arguments.callee, false);
          _this.onLoad(aEvent.currentTarget);
        }, false);
        break;
      case "xul-window-visible":
        // get window
        // let win = this._windowWatcher.activeWindow;
        // this.onVisible(win);
        // with patch!
        this.onVisible(aSubject);
        break;
    }

    this._prettyPrint();
  },

  onLoad: function(aWindow) {
    // is it a browser window? are we already tracking it?
    if (!this._isNavigator(aWindow) || aWindow.__NP)
      return;

    // assign unique identifier?
    aWindow.__NP = {
      id: "w_" + Date.now(),
      lastSelectTabIndex: -1,
      basePriority: Ci.nsISupportsPriority.PRIORITY_NORMAL
    };

    // add event listeners
    aWindow.gBrowser.addEventListener("TabOpen", this, true);
    aWindow.gBrowser.addEventListener("TabClose", this, true);
    aWindow.gBrowser.addEventListener("TabSelect", this, true);
    // aWindow.addEventListener("focus", this, true);
    // aWindow.gBrowser.addEventListener("focus", this, true);
  },

  onVisible: function(aWindow) {
    // is it a browser window?
    if (!this._isNavigator(aWindow))
      return;

    log(aWindow.__NP.toSource());

    let tabbrowser = aWindow.getBrowser();
    log(tabbrowser.contentTitle);


    let windowMediator = Cc["@mozilla.org/appshell/window-mediator;1"].
                         getService(Ci.nsIWindowMediator);
    let windowsEnum = windowMediator.getEnumerator("navigator:browser");
    
    if (this._lastFocusedWindowID) {
      let lastFocusedWindow = null;
      while (windowsEnum.hasMoreElements()) {
        let window = windowsEnum.getNext();
        if (window.__NP && window.__NP.id == this._lastFocusedWindowID)
          lastFocusedWindow = window;
      }
      this._setWindowPriority(lastFocusedWindow, "unfocused");
    }
    this._setWindowPriority(aWindow, "focused");
    this._lastFocusedWindowID = aWindow.__NP.id;


    // check if we've setup this window yet (would be via this.onLoad)
    //    if not, setTimeout
    // "domwindowopened" happens before "xul-window-visible" so we should be fine...
  },

  handleEvent: function(aEvent) {
    log(aEvent.type);
    let win = aEvent.currentTarget.ownerDocument.defaultView;
    switch (aEvent.type) {
      case "TabOpen":
        this.onTabOpen(win, aEvent.originalTarget.linkedBrowser);
        break;
      case "TabSelect":
        let browser = aEvent.originalTarget.linkedBrowser;
        this.onTabSelect(win, browser);
        break;
      case "TabClose":
        this.onTabClose(win, aEvent.originalTarget);
        break;
    }
    this._prettyPrint();
  },

  onTabClose: function(aWindow, aTab) {
    // just remove the tab from our tracking
    // Or not? There isn't really anything to do (at least not the way we're tracking now)
  },

  onTabOpen: function(aWindow, aBrowser) {
    // just set it's priority to that of the window. we'll get TabSelect after this if it's an open+select
    this._setTabPriority(aBrowser, aWindow.__NP.basePriority);
  },

  onTabSelect: function(aWindow, aBrowser) {
    log(aWindow.__NP.toSource());

    // let tabbrowser = aWindow.getBrowser();

    // log(tabbrowser.tabContainer.selectedIndex);

    let winState = (aWindow.__NP.id == this._lastFocusedWindowID) ? "focused" : "unfocused";
    log(winState);

    // BAAAAADDDDDDDDDDDD
    let _this = this;
    Array.forEach(aWindow.gBrowser.browsers, function(browser) {
      let tabFocus = (aWindow.gBrowser.selectedBrowser == browser) ? "active" : "inactive";
      _this._setTabPriority(browser, PRIORITIES[winState][tabFocus]);
    });


    // set old tab's priority down
    // i guess we'll be keeping track of that (per window)?

    // bump this tabs priority up
    // this._setTabPriority(aBrowser, PRIORITIES[winState]["active"]);
    //
  },

  _setTabPriority: function(aBrowser, aPriority) {
    let loadgroup = aBrowser.webNavigation.QueryInterface(Ci.nsIDocumentLoader).loadGroup.
                                           QueryInterface(Ci.nsISupportsPriority);
    loadgroup.priority = aPriority;
  },

  _getTabPriority: function(aBrowser) {
    return aBrowser.webNavigation.QueryInterface(Ci.nsIDocumentLoader).loadGroup.
                                  QueryInterface(Ci.nsISupportsPriority).priority;
  },

  _setWindowPriority: function(aWindow, aFocus) {
    // if we have a last focused window, then we should change it's priority
    if (aFocus == "focused" && this._lastFocusedWindowID &&
        this._lastFocusedWindowID != aWindow.__NP.id) {
      let lastFocusedWindow = null;
      let windowMediator = Cc["@mozilla.org/appshell/window-mediator;1"].
                           getService(Ci.nsIWindowMediator);
      let windowsEnum = windowMediator.getEnumerator("navigator:browser");
      while (windowsEnum.hasMoreElements()) {
        let window = windowsEnum.getNext();
        if (window.__NP && window.__NP.id == this._lastFocusedWindowID)
          lastFocusedWindow = window;
      }
      this._lastFocusedWindowID = null;
      if (lastFocusedWindow)
        this._setWindowPriority(lastFocusedWindow, "unfocused");
    }

    let _this = this;
    // set priority for each tab in the window
    Array.forEach(aWindow.gBrowser.browsers, function(browser) {
      let tabFocus = (aWindow.gBrowser.selectedBrowser == browser) ? "active" : "inactive";
      _this._setTabPriority(browser, PRIORITIES[aFocus][tabFocus]);
    });

    // set the window's base priority
    aWindow.__NP.basePriority = PRIORITIES[aFocus]["inactive"];
  },

  _isNavigator: function(aWindow) {
    return aWindow.document.documentElement.getAttribute("windowtype") == "navigator:browser";
  },

  _prettyPrint: function() {
    let _this = this;
    let lines = [];
    lines.push(">>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>");
    let windowMediator = Cc["@mozilla.org/appshell/window-mediator;1"].
                         getService(Ci.nsIWindowMediator);
    let windowsEnum = windowMediator.getEnumerator("navigator:browser");
    while (windowsEnum.hasMoreElements()) {
      let window = windowsEnum.getNext();
      let tabbrowser = window.gBrowser;

      // skip window if it's not tracked yet...
      if (!window.__NP) continue;

      lines.push(window.__NP.id + " - " + tabbrowser.contentTitle);
      Array.forEach(tabbrowser.browsers, function(browser) {
        let line = [];
        line.push((tabbrowser.selectedBrowser == browser) ? ">" : " ");
        line.push(_this._getTabPriority(browser));
        line.push(browser.currentURI.spec);
        lines.push(line.join(" "));
      });
      lines.push("*******************************************************************************");
    }
    dump(lines.join("\n") + "\n");
  }

}


function NSGetModule(aComMgr, aFileSpec) {
  return XPCOMUtils.generateModule([NetworkPrioritizer]);
}

